@@ -1,6 +1,7 @@
+//go:build windows
 // +build windows
 
-// Copyright 2019 The TCell Authors
+// Copyright 2022 The TCell Authors
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use file except in compliance with the License.
@@ -18,6 +19,9 @@ package tcell
 
 import (
 	"errors"
+	"fmt"
+	"os"
+	"strings"
 	"sync"
 	"syscall"
 	"unicode/utf16"
@@ -36,16 +40,25 @@ type cScreen struct {
 	style      Style
 	clear      bool
 	fini       bool
+	vten       bool
+	truecolor  bool
+	running    bool
 
 	w int
 	h int
 
-	oscreen consoleInfo
-	ocursor cursorInfo
-	oimode  uint32
-	oomode  uint32
-	cells   CellBuffer
-	colors  map[Color]Color
+	oscreen     consoleInfo
+	ocursor     cursorInfo
+	cursorStyle CursorStyle
+	oimode      uint32
+	oomode      uint32
+	cells       CellBuffer
+
+	finiOnce sync.Once
+
+	mouseEnabled bool
+	wg           sync.WaitGroup
+	stopQ        chan struct{}
 
 	sync.Mutex
 }
@@ -90,30 +103,35 @@ var winColors = map[Color]Color{
 	ColorWhite:   ColorWhite,
 }
 
-var k32 = syscall.NewLazyDLL("kernel32.dll")
+var (
+	k32 = syscall.NewLazyDLL("kernel32.dll")
+	u32 = syscall.NewLazyDLL("user32.dll")
+)
 
-// We have to bring in the kernel32.dll directly, so we can get access to some
-// system calls that the core Go API lacks.
+// We have to bring in the kernel32 and user32 DLLs directly, so we can get
+// access to some system calls that the core Go API lacks.
 //
 // Note that Windows appends some functions with W to indicate that wide
 // characters (Unicode) are in use.  The documentation refers to them
 // without this suffix, as the resolution is made via preprocessor.
 var (
-	procReadConsoleInput           = k32.NewProc("ReadConsoleInputW")
-	procWaitForMultipleObjects     = k32.NewProc("WaitForMultipleObjects")
-	procCreateEvent                = k32.NewProc("CreateEventW")
-	procSetEvent                   = k32.NewProc("SetEvent")
-	procGetConsoleCursorInfo       = k32.NewProc("GetConsoleCursorInfo")
-	procSetConsoleCursorInfo       = k32.NewProc("SetConsoleCursorInfo")
-	procSetConsoleCursorPosition   = k32.NewProc("SetConsoleCursorPosition")
-	procSetConsoleMode             = k32.NewProc("SetConsoleMode")
-	procGetConsoleMode             = k32.NewProc("GetConsoleMode")
-	procGetConsoleScreenBufferInfo = k32.NewProc("GetConsoleScreenBufferInfo")
-	procFillConsoleOutputAttribute = k32.NewProc("FillConsoleOutputAttribute")
-	procFillConsoleOutputCharacter = k32.NewProc("FillConsoleOutputCharacterW")
-	procSetConsoleWindowInfo       = k32.NewProc("SetConsoleWindowInfo")
-	procSetConsoleScreenBufferSize = k32.NewProc("SetConsoleScreenBufferSize")
-	procSetConsoleTextAttribute    = k32.NewProc("SetConsoleTextAttribute")
+	procReadConsoleInput            = k32.NewProc("ReadConsoleInputW")
+	procWaitForMultipleObjects      = k32.NewProc("WaitForMultipleObjects")
+	procCreateEvent                 = k32.NewProc("CreateEventW")
+	procSetEvent                    = k32.NewProc("SetEvent")
+	procGetConsoleCursorInfo        = k32.NewProc("GetConsoleCursorInfo")
+	procSetConsoleCursorInfo        = k32.NewProc("SetConsoleCursorInfo")
+	procSetConsoleCursorPosition    = k32.NewProc("SetConsoleCursorPosition")
+	procSetConsoleMode              = k32.NewProc("SetConsoleMode")
+	procGetConsoleMode              = k32.NewProc("GetConsoleMode")
+	procGetConsoleScreenBufferInfo  = k32.NewProc("GetConsoleScreenBufferInfo")
+	procFillConsoleOutputAttribute  = k32.NewProc("FillConsoleOutputAttribute")
+	procFillConsoleOutputCharacter  = k32.NewProc("FillConsoleOutputCharacterW")
+	procSetConsoleWindowInfo        = k32.NewProc("SetConsoleWindowInfo")
+	procSetConsoleScreenBufferSize  = k32.NewProc("SetConsoleScreenBufferSize")
+	procSetConsoleTextAttribute     = k32.NewProc("SetConsoleTextAttribute")
+	procGetLargestConsoleWindowSize = k32.NewProc("GetLargestConsoleWindowSize")
+	procMessageBeep                 = u32.NewProc("MessageBeep")
 )
 
 const (
@@ -121,6 +139,39 @@ const (
 	w32WaitObject0 = uintptr(0)
 )
 
+const (
+	// VT100/XTerm escapes understood by the console
+	vtShowCursor              = "\x1b[?25h"
+	vtHideCursor              = "\x1b[?25l"
+	vtCursorPos               = "\x1b[%d;%dH" // Note that it is Y then X
+	vtSgr0                    = "\x1b[0m"
+	vtBold                    = "\x1b[1m"
+	vtUnderline               = "\x1b[4m"
+	vtBlink                   = "\x1b[5m" // Not sure this is processed
+	vtReverse                 = "\x1b[7m"
+	vtSetFg                   = "\x1b[38;5;%dm"
+	vtSetBg                   = "\x1b[48;5;%dm"
+	vtSetFgRGB                = "\x1b[38;2;%d;%d;%dm" // RGB
+	vtSetBgRGB                = "\x1b[48;2;%d;%d;%dm" // RGB
+	vtCursorDefault           = "\x1b[0 q"
+	vtCursorBlinkingBlock     = "\x1b[1 q"
+	vtCursorSteadyBlock       = "\x1b[2 q"
+	vtCursorBlinkingUnderline = "\x1b[3 q"
+	vtCursorSteadyUnderline   = "\x1b[4 q"
+	vtCursorBlinkingBar       = "\x1b[5 q"
+	vtCursorSteadyBar         = "\x1b[6 q"
+)
+
+var vtCursorStyles = map[CursorStyle]string{
+	CursorStyleDefault:           vtCursorDefault,
+	CursorStyleBlinkingBlock:     vtCursorBlinkingBlock,
+	CursorStyleSteadyBlock:       vtCursorSteadyBlock,
+	CursorStyleBlinkingUnderline: vtCursorBlinkingUnderline,
+	CursorStyleSteadyUnderline:   vtCursorSteadyUnderline,
+	CursorStyleBlinkingBar:       vtCursorBlinkingBar,
+	CursorStyleSteadyBar:         vtCursorSteadyBar,
+}
+
 // NewConsoleScreen returns a Screen for the Windows console associated
 // with the current process.  The Screen makes use of the Windows Console
 // API to display content and read events.
@@ -140,20 +191,28 @@ func (s *cScreen) Init() error {
 	s.in = in
 	out, e := syscall.Open("CONOUT$", syscall.O_RDWR, 0)
 	if e != nil {
-		syscall.Close(s.in)
+		_ = syscall.Close(s.in)
 		return e
 	}
 	s.out = out
 
-	cf, _, e := procCreateEvent.Call(
-		uintptr(0),
-		uintptr(1),
-		uintptr(0),
-		uintptr(0))
-	if cf == uintptr(0) {
-		return e
+	s.truecolor = true
+
+	// ConEmu handling of colors and scrolling when in terminal
+	// mode is extremely problematic at the best.  The color
+	// palette will scroll even though characters do not, when
+	// emitting stuff for the last character.  In the future we
+	// might change this to look at specific versions of ConEmu
+	// if they fix the bug.
+	if os.Getenv("ConEmuPID") != "" {
+		s.truecolor = false
+	}
+	switch os.Getenv("TCELL_TRUECOLOR") {
+	case "disable":
+		s.truecolor = false
+	case "enable":
+		s.truecolor = true
 	}
-	s.cancelflag = syscall.Handle(cf)
 
 	s.Lock()
 
@@ -167,14 +226,27 @@ func (s *cScreen) Init() error {
 	s.resize()
 
 	s.fini = false
-	s.setInMode(modeResizeEn)
-	s.setOutMode(0)
-	s.clearScreen(s.style)
-	s.hideCursor()
-	s.Unlock()
-	go s.scanInput()
+	s.setInMode(modeResizeEn | modeExtendFlg)
 
-	return nil
+	// 24-bit color is opt-in for now, because we can't figure out
+	// to make it work consistently.
+	if s.truecolor {
+		s.setOutMode(modeVtOutput | modeNoAutoNL | modeCookedOut)
+		var om uint32
+		s.getOutMode(&om)
+		if om&modeVtOutput == modeVtOutput {
+			s.vten = true
+		} else {
+			s.truecolor = false
+			s.setOutMode(0)
+		}
+	} else {
+		s.setOutMode(0)
+	}
+
+	s.Unlock()
+
+	return s.engage()
 }
 
 func (s *cScreen) CharacterSet() string {
@@ -182,39 +254,103 @@ func (s *cScreen) CharacterSet() string {
 	return "UTF-16LE"
 }
 
-func (s *cScreen) EnableMouse() {
-	s.setInMode(modeResizeEn | modeMouseEn | modeExtndFlg)
+func (s *cScreen) EnableMouse(...MouseFlags) {
+	s.Lock()
+	s.mouseEnabled = true
+	s.enableMouse(true)
+	s.Unlock()
 }
 
 func (s *cScreen) DisableMouse() {
-	s.setInMode(modeResizeEn)
+	s.Lock()
+	s.mouseEnabled = false
+	s.enableMouse(false)
+	s.Unlock()
 }
 
+func (s *cScreen) enableMouse(on bool) {
+	if on {
+		s.setInMode(modeResizeEn | modeMouseEn | modeExtendFlg)
+	} else {
+		s.setInMode(modeResizeEn | modeExtendFlg)
+	}
+}
+
+// Windows lacks bracketed paste (for now)
+
+func (s *cScreen) EnablePaste() {}
+
+func (s *cScreen) DisablePaste() {}
+
 func (s *cScreen) Fini() {
+	s.disengage()
+}
+
+func (s *cScreen) disengage() {
 	s.Lock()
-	s.style = StyleDefault
-	s.curx = -1
-	s.cury = -1
-	s.fini = true
+	if !s.running {
+		s.Unlock()
+		return
+	}
+	s.running = false
+	stopQ := s.stopQ
+	_, _, _ = procSetEvent.Call(uintptr(s.cancelflag))
+	close(stopQ)
 	s.Unlock()
 
-	s.setCursorInfo(&s.ocursor)
+	s.wg.Wait()
+
+	if s.vten {
+		s.emitVtString(vtCursorStyles[CursorStyleDefault])
+	}
 	s.setInMode(s.oimode)
 	s.setOutMode(s.oomode)
 	s.setBufferSize(int(s.oscreen.size.x), int(s.oscreen.size.y))
-	s.clearScreen(StyleDefault)
-	s.setCursorPos(0, 0)
-	procSetConsoleTextAttribute.Call(
+	s.clearScreen(StyleDefault, false)
+	s.setCursorPos(0, 0, false)
+	s.setCursorInfo(&s.ocursor)
+	_, _, _ = procSetConsoleTextAttribute.Call(
 		uintptr(s.out),
 		uintptr(s.mapStyle(StyleDefault)))
+}
 
-	close(s.quit)
-	procSetEvent.Call(uintptr(s.cancelflag))
-	// Block until scanInput returns; this prevents a race condition on Win 8+
-	// which causes syscall.Close to block until another keypress is read.
-	<-s.scandone
-	syscall.Close(s.in)
-	syscall.Close(s.out)
+func (s *cScreen) engage() error {
+	s.Lock()
+	defer s.Unlock()
+	if s.running {
+		return errors.New("already engaged")
+	}
+	s.stopQ = make(chan struct{})
+	cf, _, e := procCreateEvent.Call(
+		uintptr(0),
+		uintptr(1),
+		uintptr(0),
+		uintptr(0))
+	if cf == uintptr(0) {
+		return e
+	}
+	s.running = true
+	s.cancelflag = syscall.Handle(cf)
+	s.enableMouse(s.mouseEnabled)
+
+	if s.vten {
+		s.setOutMode(modeVtOutput | modeNoAutoNL | modeCookedOut)
+	} else {
+		s.setOutMode(0)
+	}
+
+	s.clearScreen(s.style, s.vten)
+	s.hideCursor()
+
+	s.cells.Invalidate()
+	s.hideCursor()
+	s.resize()
+	s.draw()
+	s.doCursor()
+
+	s.wg.Add(1)
+	go s.scanInput(s.stopQ)
+	return nil
 }
 
 func (s *cScreen) PostEventWait(ev Event) {
@@ -230,15 +366,39 @@ func (s *cScreen) PostEvent(ev Event) error {
 	}
 }
 
+func (s *cScreen) ChannelEvents(ch chan<- Event, quit <-chan struct{}) {
+	defer close(ch)
+	for {
+		select {
+		case <-quit:
+			return
+		case <-s.stopQ:
+			return
+		case ev := <-s.evch:
+			select {
+			case <-quit:
+				return
+			case <-s.stopQ:
+				return
+			case ch <- ev:
+			}
+		}
+	}
+}
+
 func (s *cScreen) PollEvent() Event {
 	select {
-	case <-s.quit:
+	case <-s.stopQ:
 		return nil
 	case ev := <-s.evch:
 		return ev
 	}
 }
 
+func (s *cScreen) HasPendingEvent() bool {
+	return len(s.evch) > 0
+}
+
 type cursorInfo struct {
 	size    uint32
 	visible uint32
@@ -261,12 +421,26 @@ type rect struct {
 	bottom int16
 }
 
+func (s *cScreen) emitVtString(vs string) {
+	esc := utf16.Encode([]rune(vs))
+	_ = syscall.WriteConsole(s.out, &esc[0], uint32(len(esc)), nil, nil)
+}
+
 func (s *cScreen) showCursor() {
-	s.setCursorInfo(&cursorInfo{size: 100, visible: 1})
+	if s.vten {
+		s.emitVtString(vtShowCursor)
+		s.emitVtString(vtCursorStyles[s.cursorStyle])
+	} else {
+		s.setCursorInfo(&cursorInfo{size: 100, visible: 1})
+	}
 }
 
 func (s *cScreen) hideCursor() {
-	s.setCursorInfo(&cursorInfo{size: 1, visible: 0})
+	if s.vten {
+		s.emitVtString(vtHideCursor)
+	} else {
+		s.setCursorInfo(&cursorInfo{size: 1, visible: 0})
+	}
 }
 
 func (s *cScreen) ShowCursor(x, y int) {
@@ -279,13 +453,24 @@ func (s *cScreen) ShowCursor(x, y int) {
 	s.Unlock()
 }
 
+func (s *cScreen) SetCursorStyle(cs CursorStyle) {
+	s.Lock()
+	if !s.fini {
+		if _, ok := vtCursorStyles[cs]; ok {
+			s.cursorStyle = cs
+			s.doCursor()
+		}
+	}
+	s.Unlock()
+}
+
 func (s *cScreen) doCursor() {
 	x, y := s.curx, s.cury
 
 	if x < 0 || y < 0 || x >= s.w || y >= s.h {
 		s.hideCursor()
 	} else {
-		s.setCursorPos(x, y)
+		s.setCursorPos(x, y, s.vten)
 		s.showCursor()
 	}
 }
@@ -294,11 +479,6 @@ func (s *cScreen) HideCursor() {
 	s.ShowCursor(-1, -1)
 }
 
-type charInfo struct {
-	ch   uint16
-	attr uint16
-}
-
 type inputRecord struct {
 	typ  uint16
 	_    uint16
@@ -309,8 +489,8 @@ const (
 	keyEvent    uint16 = 1
 	mouseEvent  uint16 = 2
 	resizeEvent uint16 = 4
-	menuEvent   uint16 = 8  // don't use
-	focusEvent  uint16 = 16 // don't use
+	// menuEvent   uint16 = 8  // don't use
+	// focusEvent  uint16 = 16 // don't use
 )
 
 type mouseRecord struct {
@@ -322,10 +502,10 @@ type mouseRecord struct {
 }
 
 const (
-	mouseDoubleClick uint32 = 0x2
-	mouseHWheeled    uint32 = 0x8
-	mouseVWheeled    uint32 = 0x4
-	mouseMoved       uint32 = 0x1
+	mouseHWheeled uint32 = 0x8
+	mouseVWheeled uint32 = 0x4
+	// mouseDoubleClick uint32 = 0x2
+	// mouseMoved       uint32 = 0x1
 )
 
 type resizeRecord struct {
@@ -412,6 +592,8 @@ var vkKeys = map[uint16]Key{
 	vkInsert: KeyInsert,
 	vkDelete: KeyDelete,
 	vkHelp:   KeyHelp,
+	vkEscape: KeyEscape,
+	vkSpace:  ' ',
 	vkF1:     KeyF1,
 	vkF2:     KeyF2,
 	vkF3:     KeyF3,
@@ -566,10 +748,10 @@ func (s *cScreen) getConsoleInput() error {
 				for krec.repeat > 0 {
 					// convert shift+tab to backtab
 					if mod2mask(krec.mod) == ModShift && krec.ch == vkTab {
-						s.PostEvent(NewEventKey(KeyBacktab, 0,
+						s.PostEventWait(NewEventKey(KeyBacktab, 0,
 							ModNone))
 					} else {
-						s.PostEvent(NewEventKey(KeyRune, rune(krec.ch),
+						s.PostEventWait(NewEventKey(KeyRune, rune(krec.ch),
 							mod2mask(krec.mod)))
 					}
 					krec.repeat--
@@ -582,7 +764,7 @@ func (s *cScreen) getConsoleInput() error {
 				return nil
 			}
 			for krec.repeat > 0 {
-				s.PostEvent(NewEventKey(key, rune(krec.ch),
+				s.PostEventWait(NewEventKey(key, rune(krec.ch),
 					mod2mask(krec.mod)))
 				krec.repeat--
 			}
@@ -596,14 +778,14 @@ func (s *cScreen) getConsoleInput() error {
 			mrec.flags = getu32(rec.data[12:])
 			btns := mrec2btns(mrec.btns, mrec.flags)
 			// we ignore double click, events are delivered normally
-			s.PostEvent(NewEventMouse(int(mrec.x), int(mrec.y), btns,
+			s.PostEventWait(NewEventMouse(int(mrec.x), int(mrec.y), btns,
 				mod2mask(mrec.mod)))
 
 		case resizeEvent:
 			var rrec resizeRecord
 			rrec.x = geti16(rec.data[0:])
 			rrec.y = geti16(rec.data[2:])
-			s.PostEvent(NewEventResize(int(rrec.x), int(rrec.y)))
+			s.PostEventWait(NewEventResize(int(rrec.x), int(rrec.y)))
 
 		default:
 		}
@@ -614,17 +796,25 @@ func (s *cScreen) getConsoleInput() error {
 	return nil
 }
 
-func (s *cScreen) scanInput() {
+func (s *cScreen) scanInput(stopQ chan struct{}) {
+	defer s.wg.Done()
 	for {
+		select {
+		case <-stopQ:
+			return
+		default:
+		}
 		if e := s.getConsoleInput(); e != nil {
-			close(s.scandone)
 			return
 		}
 	}
 }
 
-// Windows console can display 8 characters, in either low or high intensity
 func (s *cScreen) Colors() int {
+	if s.vten {
+		return 1 << 24
+	}
+	// Windows console can display 8 colors, in either low or high intensity
 	return 16
 }
 
@@ -670,10 +860,10 @@ func (s *cScreen) mapStyle(style Style) uint16 {
 	f, b, a := style.Decompose()
 	fa := s.oscreen.attrs & 0xf
 	ba := (s.oscreen.attrs) >> 4 & 0xf
-	if f != ColorDefault {
+	if f != ColorDefault && f != ColorReset {
 		fa = mapColor2RGB(f)
 	}
-	if b != ColorDefault {
+	if b != ColorDefault && b != ColorReset {
 		ba = mapColor2RGB(b)
 	}
 	var attr uint16
@@ -682,10 +872,10 @@ func (s *cScreen) mapStyle(style Style) uint16 {
 	// views.
 	if a&AttrReverse != 0 {
 		attr = ba
-		attr |= (fa << 4)
+		attr |= fa << 4
 	} else {
 		attr = fa
-		attr |= (ba << 4)
+		attr |= ba << 4
 	}
 	if a&AttrBold != 0 {
 		attr |= 0x8
@@ -709,19 +899,53 @@ func (s *cScreen) SetCell(x, y int, style Style, ch ...rune) {
 	}
 }
 
-func (s *cScreen) SetContent(x, y int, mainc rune, combc []rune, style Style) {
+func (s *cScreen) SetContent(x, y int, primary rune, combining []rune, style Style) {
 	s.Lock()
 	if !s.fini {
-		s.cells.SetContent(x, y, mainc, combc, style)
+		s.cells.SetContent(x, y, primary, combining, style)
 	}
 	s.Unlock()
 }
 
 func (s *cScreen) GetContent(x, y int) (rune, []rune, Style, int) {
 	s.Lock()
-	mainc, combc, style, width := s.cells.GetContent(x, y)
+	primary, combining, style, width := s.cells.GetContent(x, y)
 	s.Unlock()
-	return mainc, combc, style, width
+	return primary, combining, style, width
+}
+
+func (s *cScreen) sendVtStyle(style Style) {
+	esc := &strings.Builder{}
+
+	fg, bg, attrs := style.Decompose()
+
+	esc.WriteString(vtSgr0)
+
+	if attrs&(AttrBold|AttrDim) == AttrBold {
+		esc.WriteString(vtBold)
+	}
+	if attrs&AttrBlink != 0 {
+		esc.WriteString(vtBlink)
+	}
+	if attrs&AttrUnderline != 0 {
+		esc.WriteString(vtUnderline)
+	}
+	if attrs&AttrReverse != 0 {
+		esc.WriteString(vtReverse)
+	}
+	if fg.IsRGB() {
+		r, g, b := fg.RGB()
+		_, _ = fmt.Fprintf(esc, vtSetFgRGB, r, g, b)
+	} else if fg.Valid() {
+		_, _ = fmt.Fprintf(esc, vtSetFg, fg&0xff)
+	}
+	if bg.IsRGB() {
+		r, g, b := bg.RGB()
+		_, _ = fmt.Fprintf(esc, vtSetBgRGB, r, g, b)
+	} else if bg.Valid() {
+		_, _ = fmt.Fprintf(esc, vtSetBg, bg&0xff)
+	}
+	s.emitVtString(esc.String())
 }
 
 func (s *cScreen) writeString(x, y int, style Style, ch []uint16) {
@@ -729,31 +953,35 @@ func (s *cScreen) writeString(x, y int, style Style, ch []uint16) {
 	if len(ch) == 0 {
 		return
 	}
-	nw := uint32(len(ch))
-	procSetConsoleTextAttribute.Call(
-		uintptr(s.out),
-		uintptr(s.mapStyle(style)))
-	s.setCursorPos(x, y)
-	syscall.WriteConsole(s.out, &ch[0], nw, &nw, nil)
+	s.setCursorPos(x, y, s.vten)
+
+	if s.vten {
+		s.sendVtStyle(style)
+	} else {
+		_, _, _ = procSetConsoleTextAttribute.Call(
+			uintptr(s.out),
+			uintptr(s.mapStyle(style)))
+	}
+	_ = syscall.WriteConsole(s.out, &ch[0], uint32(len(ch)), nil, nil)
 }
 
 func (s *cScreen) draw() {
 	// allocate a scratch line bit enough for no combining chars.
-	// if you have combining characters, you may pay for extra allocs.
+	// if you have combining characters, you may pay for extra allocations.
 	if s.clear {
-		s.clearScreen(s.style)
+		s.clearScreen(s.style, s.vten)
 		s.clear = false
 		s.cells.Invalidate()
 	}
 	buf := make([]uint16, 0, s.w)
 	wcs := buf[:]
-	lstyle := Style(-1) // invalid attribute
+	lstyle := styleInvalid
 
 	lx, ly := -1, -1
 	ra := make([]rune, 1)
 
-	for y := 0; y < int(s.h); y++ {
-		for x := 0; x < int(s.w); x++ {
+	for y := 0; y < s.h; y++ {
+		for x := 0; x < s.w; x++ {
 			mainc, combc, style, width := s.cells.GetContent(x, y)
 			dirty := s.cells.Dirty(x, y)
 			if style == StyleDefault {
@@ -766,7 +994,7 @@ func (s *cScreen) draw() {
 				// cells, or because we need to change styles
 				s.writeString(lx, ly, lstyle, wcs)
 				wcs = buf[0:0]
-				lstyle = Style(-1)
+				lstyle = StyleDefault
 				if !dirty {
 					continue
 				}
@@ -793,7 +1021,7 @@ func (s *cScreen) draw() {
 		}
 		s.writeString(lx, ly, lstyle, wcs)
 		wcs = buf[0:0]
-		lstyle = Style(-1)
+		lstyle = styleInvalid
 	}
 }
 
@@ -829,31 +1057,37 @@ type consoleInfo struct {
 }
 
 func (s *cScreen) getConsoleInfo(info *consoleInfo) {
-	procGetConsoleScreenBufferInfo.Call(
+	_, _, _ = procGetConsoleScreenBufferInfo.Call(
 		uintptr(s.out),
 		uintptr(unsafe.Pointer(info)))
 }
 
 func (s *cScreen) getCursorInfo(info *cursorInfo) {
-	procGetConsoleCursorInfo.Call(
+	_, _, _ = procGetConsoleCursorInfo.Call(
 		uintptr(s.out),
 		uintptr(unsafe.Pointer(info)))
 }
 
 func (s *cScreen) setCursorInfo(info *cursorInfo) {
-	procSetConsoleCursorInfo.Call(
+	_, _, _ = procSetConsoleCursorInfo.Call(
 		uintptr(s.out),
 		uintptr(unsafe.Pointer(info)))
+
 }
 
-func (s *cScreen) setCursorPos(x, y int) {
-	procSetConsoleCursorPosition.Call(
-		uintptr(s.out),
-		coord{int16(x), int16(y)}.uintptr())
+func (s *cScreen) setCursorPos(x, y int, vtEnable bool) {
+	if vtEnable {
+		// Note that the string is Y first.  Origin is 1,1.
+		s.emitVtString(fmt.Sprintf(vtCursorPos, y+1, x+1))
+	} else {
+		_, _, _ = procSetConsoleCursorPosition.Call(
+			uintptr(s.out),
+			coord{int16(x), int16(y)}.uintptr())
+	}
 }
 
 func (s *cScreen) setBufferSize(x, y int) {
-	procSetConsoleScreenBufferSize.Call(
+	_, _, _ = procSetConsoleScreenBufferSize.Call(
 		uintptr(s.out),
 		coord{int16(x), int16(y)}.uintptr())
 }
@@ -866,6 +1100,37 @@ func (s *cScreen) Size() (int, int) {
 	return w, h
 }
 
+func (s *cScreen) SetSize(w, h int) {
+	xy, _, _ := procGetLargestConsoleWindowSize.Call(uintptr(s.out))
+
+	// xy is little endian packed
+	y := int(xy >> 16)
+	x := int(xy & 0xffff)
+
+	if x == 0 || y == 0 {
+		return
+	}
+
+	// This is a hacky workaround for Windows Terminal.
+	// Essentially Windows Terminal (Windows 11) does not support application
+	// initiated resizing.  To detect this, we look for an extremely large size
+	// for the maximum width.  If it is > 500, then this is almost certainly
+	// Windows Terminal, and won't support this.  (Note that the legacy console
+	// does support application resizing.)
+	if x >= 500 {
+		return
+	}
+
+	s.setBufferSize(x, y)
+	r := rect{0, 0, int16(w - 1), int16(h - 1)}
+	_, _, _ = procSetConsoleWindowInfo.Call(
+		uintptr(s.out),
+		uintptr(1),
+		uintptr(unsafe.Pointer(&r)))
+
+	s.resize()
+}
+
 func (s *cScreen) resize() {
 	info := consoleInfo{}
 	s.getConsoleInfo(&info)
@@ -884,12 +1149,11 @@ func (s *cScreen) resize() {
 	s.setBufferSize(w, h)
 
 	r := rect{0, 0, int16(w - 1), int16(h - 1)}
-	procSetConsoleWindowInfo.Call(
+	_, _, _ = procSetConsoleWindowInfo.Call(
 		uintptr(s.out),
 		uintptr(1),
 		uintptr(unsafe.Pointer(&r)))
-
-	s.PostEvent(NewEventResize(w, h))
+	_ = s.PostEvent(NewEventResize(w, h))
 }
 
 func (s *cScreen) Clear() {
@@ -905,63 +1169,73 @@ func (s *cScreen) Fill(r rune, style Style) {
 	s.Unlock()
 }
 
-func (s *cScreen) clearScreen(style Style) {
-	pos := coord{0, 0}
-	attr := s.mapStyle(style)
-	x, y := s.w, s.h
-	scratch := uint32(0)
-	count := uint32(x * y)
+func (s *cScreen) clearScreen(style Style, vtEnable bool) {
+	if vtEnable {
+		s.sendVtStyle(style)
+		row := strings.Repeat(" ", s.w)
+		for y := 0; y < s.h; y++ {
+			s.setCursorPos(0, y, vtEnable)
+			s.emitVtString(row)
+		}
+		s.setCursorPos(0, 0, vtEnable)
 
-	procFillConsoleOutputAttribute.Call(
-		uintptr(s.out),
-		uintptr(attr),
-		uintptr(count),
-		pos.uintptr(),
-		uintptr(unsafe.Pointer(&scratch)))
-	procFillConsoleOutputCharacter.Call(
-		uintptr(s.out),
-		uintptr(' '),
-		uintptr(count),
-		pos.uintptr(),
-		uintptr(unsafe.Pointer(&scratch)))
+	} else {
+		pos := coord{0, 0}
+		attr := s.mapStyle(style)
+		x, y := s.w, s.h
+		scratch := uint32(0)
+		count := uint32(x * y)
+
+		_, _, _ = procFillConsoleOutputAttribute.Call(
+			uintptr(s.out),
+			uintptr(attr),
+			uintptr(count),
+			pos.uintptr(),
+			uintptr(unsafe.Pointer(&scratch)))
+		_, _, _ = procFillConsoleOutputCharacter.Call(
+			uintptr(s.out),
+			uintptr(' '),
+			uintptr(count),
+			pos.uintptr(),
+			uintptr(unsafe.Pointer(&scratch)))
+	}
 }
 
 const (
-	modeExtndFlg uint32 = 0x0080
-	modeMouseEn  uint32 = 0x0010
-	modeResizeEn uint32 = 0x0008
-	modeWrapEOL  uint32 = 0x0002
-	modeCooked   uint32 = 0x0001
+	// Input modes
+	modeExtendFlg uint32 = 0x0080
+	modeMouseEn          = 0x0010
+	modeResizeEn         = 0x0008
+	// modeCooked          = 0x0001
+	// modeVtInput         = 0x0200
+
+	// Output modes
+	modeCookedOut uint32 = 0x0001
+	modeVtOutput         = 0x0004
+	modeNoAutoNL         = 0x0008
+	// modeWrapEOL          = 0x0002
 )
 
-func (s *cScreen) setInMode(mode uint32) error {
-	rv, _, err := procSetConsoleMode.Call(
+func (s *cScreen) setInMode(mode uint32) {
+	_, _, _ = procSetConsoleMode.Call(
 		uintptr(s.in),
 		uintptr(mode))
-	if rv == 0 {
-		return err
-	}
-	return nil
 }
 
-func (s *cScreen) setOutMode(mode uint32) error {
-	rv, _, err := procSetConsoleMode.Call(
+func (s *cScreen) setOutMode(mode uint32) {
+	_, _, _ = procSetConsoleMode.Call(
 		uintptr(s.out),
 		uintptr(mode))
-	if rv == 0 {
-		return err
-	}
-	return nil
 }
 
 func (s *cScreen) getInMode(v *uint32) {
-	procGetConsoleMode.Call(
+	_, _, _ = procGetConsoleMode.Call(
 		uintptr(s.in),
 		uintptr(unsafe.Pointer(v)))
 }
 
 func (s *cScreen) getOutMode(v *uint32) {
-	procGetConsoleMode.Call(
+	_, _, _ = procGetConsoleMode.Call(
 		uintptr(s.out),
 		uintptr(unsafe.Pointer(v)))
 }
@@ -974,15 +1248,15 @@ func (s *cScreen) SetStyle(style Style) {
 
 // No fallback rune support, since we have Unicode.  Yay!
 
-func (s *cScreen) RegisterRuneFallback(r rune, subst string) {
+func (s *cScreen) RegisterRuneFallback(_ rune, _ string) {
 }
 
-func (s *cScreen) UnregisterRuneFallback(r rune) {
+func (s *cScreen) UnregisterRuneFallback(_ rune) {
 }
 
-func (s *cScreen) CanDisplay(r rune, checkFallbacks bool) bool {
+func (s *cScreen) CanDisplay(_ rune, _ bool) bool {
 	// We presume we can display anything -- we're Unicode.
-	// (Sadly this not precisely true.  Combinings are especially
+	// (Sadly this not precisely true.  Combining characters are especially
 	// poorly supported under Windows.)
 	return true
 }
@@ -1031,3 +1305,25 @@ func (s *cScreen) HasKey(k Key) bool {
 
 	return valid[k]
 }
+
+func (s *cScreen) Beep() error {
+	// A simple beep. If the sound card is not available, the sound is generated
+	// using the speaker.
+	//
+	// Reference:
+	// https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-messagebeep
+	const simpleBeep = 0xffffffff
+	if rv, _, err := procMessageBeep.Call(simpleBeep); rv == 0 {
+		return err
+	}
+	return nil
+}
+
+func (s *cScreen) Suspend() error {
+	s.disengage()
+	return nil
+}
+
+func (s *cScreen) Resume() error {
+	return s.engage()
+}
